/*
 * @file ScanEngine.cpp
 * @brief The Main Worker Class for the RPU
 *
 * The Main Worker Class for the RPU which provides the "Runner".
 *
 * @copyright redheads.de
 * @date 2020-03-29
 * @author guenter.woigk@redheads.de
 */


#include "FatalError.h"
#include "ScanEngine.h"
#include "ILogger.hpp"
#include "IPrintServer.h"
#include "ILeds.h"
#include "settings.h"
#include "rpuconfiguration.h"
#ifdef SCAN_ENGINE_INCLUDE_DEBUG_CONSOLE
#include "IUart.h"
#endif

static inline float sin(float a) noexcept { return sinf(a); }
static inline float cos(float a) noexcept { return cosf(a); }

namespace ScanEngine
{

#if defined LASER_SCANNER_VISIBLE_JUMPS
bool laser_scanner_visible_jumps = false;
#endif

#ifdef SCAN_ENGINE_INCLUDE_DEBUG_CONSOLE
#undef LOGTRACE
#define LOGTRACE(TEXT)	do { if (logger->loglevel >= LogLevel::Trace) console->puts(TEXT); } while(0)
#endif

static const FLOAT pi = 3.1415926538f;

// array of powerOFF and powerON values
//   which can be used to construct power[10] series
//   with all ON, all OFF, OFF going ON and ON going OFF at any position
//   but no OFF-ON-OFF or ON-OFF-ON sequence
static const pwr_t power[30] =
{
#define M pwr_max
		0,0,0,0,0,0,0,0,0,0,
		M,M,M,M,M,M,M,M,M,M,
		0,0,0,0,0,0,0,0,0,0,
#undef M
};
static const pwr_t* powerOn  = power+10;
static const pwr_t* powerOff = power+20;


// my self-supplied ctor for LaserPowerData:
template<>
inline LaserPowerData::LaserPowerData(const PlaneData& q) noexcept :
	m_laserPower(q.m_laserPower),
	m_laserFrequency(q.m_laserFrequency),
	m_wobbleType(q.m_wobbleType),
	m_wobbleAmplitude(q.m_wobbleAmplitude),
	m_wobbleFrequency(q.m_wobbleFrequency),
	m_co2StandByPulsePeriod_usec(q.m_co2StandByPulsePeriod_usec),
	m_co2StandByPulseLength_usec(q.m_co2StandByPulseLength_usec),
	m_pulseLength_usec(q.m_pulseLength),
	m_firstPulseLength_usec(q.m_firstPulse)
{
}

ScanEngine::ScanEngine(IPrintServer& printserver, QueuedLaserScanner& laserscanner, const SharedMemory::RpuConfiguration& rpu_configuration) :
	m_PrintServer(printserver),
	m_LaserScanner(laserscanner),
	m_scannerFrequency(rpu_configuration.m_scannerFrequency),
	m_maxMarkSpeed_mpss(rpu_configuration.m_scanFieldWidth / rpu_configuration.m_scannerFrequency * rpu_configuration.m_maxScanSpeed),
	m_minMarkSpeed_mpss(m_maxMarkSpeed_mpss / 100),
	m_maxJumpSpeed_mpss(m_maxMarkSpeed_mpss * 5),
	m_minJumpSpeed_mpss(m_minMarkSpeed_mpss),
	m_minJumpDelay_usec(rpu_configuration.m_minJumpDelay_usec),
	m_minMarkDelay_usec(rpu_configuration.m_minMarkDelay_usec),
	m_minPolygonDelay_usec(rpu_configuration.m_minPolygonDelay_usec),
	m_minLaserOnDelay_usec(rpu_configuration.m_minLaserOnDelay_usec),
	m_minLaserOffDelay_usec(rpu_configuration.m_minLaserOffDelay_usec),
	m_maxJumpDelay_usec(rpu_configuration.m_maxJumpDelay_usec),
	m_maxMarkDelay_usec(rpu_configuration.m_maxMarkDelay_usec),
	m_maxPolygonDelay_usec(rpu_configuration.m_maxPolygonDelay_usec),
	m_maxLaserOnDelay_usec(rpu_configuration.m_maxLaserOnDelay_usec),
	m_maxLaserOffDelay_usec(rpu_configuration.m_maxLaserOffDelay_usec),
	m_currentPoint()
{
	LOGDEBUG("ctor ScanEngine");
}


void ScanEngine::setPlaneData(const PlaneData& q)
{
	LOGTRACE("P");

	////const FLOAT cMinStepsPerScanfield = cScannerFrequency * 120mm / 15m/s;
	////const FLOAT cMinStepsPerScanfield = cMinStepsPerScanfield * 1;		// physical limit
	//const FLOAT cMaxStepsPerScanfield = cMinStepsPerScanfield * 100;		// sanity limit
	//const FLOAT cMinStepsPerScanfieldForJump = cMinStepsPerScanfield / 5;	// sanity limit
    //
	//const FLOAT maxMarkSpeed_mpss = m_LaserScanner.getScanField().width() / cMinStepsPerScanfield;
	//const FLOAT minMarkSpeed_mpss = m_LaserScanner.getScanField().width() / cMaxStepsPerScanfield;
	//const FLOAT maxJumpSpeed_mpss = m_LaserScanner.getScanField().width() / cMinStepsPerScanfieldForJump;
	//const FLOAT minJumpSpeed_mpss = minMarkSpeed_mpss;

    // scan speed for marking (laser on) [m/scannerstep]
    m_markSpeed_mpss = minmax( m_minMarkSpeed_mpss, q.m_markVelocity / m_scannerFrequency, m_maxMarkSpeed_mpss );

    // scan speed for jumping (laser off) [m/scannerstep]
    m_jumpSpeed_mpss = minmax( m_minJumpSpeed_mpss, q.m_jumpVelocity / m_scannerFrequency, m_maxJumpSpeed_mpss) ;

    // delay: start of mark vector -> laser on [µs]
    m_laserOnDelay_usec = minmax(m_minLaserOnDelay_usec, q.m_laserOnDelay, m_maxLaserOnDelay_usec);

    // delay: end of mark vector -> laser off [µs]
    m_laserOffDelay_usec = minmax(m_minLaserOffDelay_usec, q.m_laserOffDelay, m_maxLaserOffDelay_usec);

    // delay: end of jump vector -> start of marking [µs]
    m_jumpEndDelay_usec = minmax(m_minJumpDelay_usec, q.m_jumpDelay, m_maxJumpDelay_usec);

    // delay: end of mark vector -> start of jump [µs]
    m_markEndDelay_usec = minmax(m_minMarkDelay_usec, q.m_markDelay, m_maxMarkDelay_usec);

    // delay: the delay during marking when switching the current markline (polyline) [µs]
    m_markMiddleDelay_usec = minmax(m_minPolygonDelay_usec, q.m_polygonDelay, m_maxPolygonDelay_usec);

    // spot size for filling algorithms [µm] //TODO wahrscheinlich nicht benötigt
    m_spotSize_um = q.m_spotSize;
}

void ScanEngine::setPlaneData(uint plane)
{
	//char buffer[30];
	//sprintf(buffer, "Planedata: %d", plane);
	//LOGINFO(buffer);

	const PlaneData& planeData = m_PrintServer.getPlaneData(plane);
	m_LaserScanner.setLaserPower(LaserPowerData(planeData));
	setPlaneData(planeData);
}

static//helper
Point movedTowardsForJump (Point p, Point dest, FLOAT speed)
{
	// move point p towards dest with speed
	// speed is handled separately for x and y
	// this is suitable for jump, not draw

	Dist dist = dest-p;
	if (dist.dx > 0)
	{
		p.x += min(dist.dx,+speed);
	}
	else
	{
		p.x += max(dist.dx,-speed);
	}
	if (dist.dy > 0)
	{
		p.y += min(dist.dy,+speed);
	}
	else
	{
		p.y += max(dist.dy,-speed);
	}

	return p;
}

static//helper
Point movedTowardsForDraw (Point p, Point dest, FLOAT speed)
{
	// move point p towards dest with speed
	// speed is handled in direction of line
	// this is suitable for draw

	Dist line = dest - p;
	FLOAT line_length = line.length();		// uses sqrt

	if (speed >= line_length)
	{
		return dest;
	}
	else
	{
		return p + line * (speed / line_length);
	}
}

void ScanEngine::jumpToPoint(const Rect& bbox, const Point& dest)
{
	// jump to dest
	// wait for bbox
	// wait m_jumpDelay

	LOGTRACE("\n");

	// jump to dest
	while (m_currentPoint != dest)
	{
		LOGTRACE("j");
		Point p = movedTowardsForJump(m_currentPoint,dest,m_jumpSpeed_mpss);
		m_LaserScanner.addPoint(p,powerOff); // note: point will be limited to scan field by LUT
		m_currentPoint = p;
	}

	int delay = m_jumpEndDelay_usec;

	// wait for bbox

	for(uint i=0;;i++)
	{
		Rect sbox = m_LaserScanner.getScanField();
		Rect dbox = m_LaserScanner.scanfieldForLabel(bbox);
		if (sbox.encloses(dbox)) break;
		assert(sbox.width()>0);

		if (logger->loglevel >= LogLevel::Trace)
		{
			uint cnt = m_LaserScanner.getAvail();
			Dist pos = m_LaserScanner.getEncoderDistance();
			Dist speed = m_LaserScanner.getEncoderDistancePerScannerStep();
			Dist dpos = m_LaserScanner.getEstimatedEncoderDistance();

			if (i==0) console->printf("\nsbox = [%f, %f]", sbox.left, sbox.right);
			if (i==0) console->printf("\nbbox = [%f, %f]", bbox.left, bbox.right);
			if (i==0) console->printf("\n encoder speed = %f ", speed.dx);
			if (i<=9) console->printf("\ndbox = [%f, %f]", dbox.left, dbox.right);
			if (i<=9) console->printf("\n encoder dist. = %f ", pos.dx);
			if (i<=9) console->printf("\n encoder dist. = %f after %u steps. ", dpos.dx, cnt);
		}

		LOGTRACE("W");           // uppercase W: wait for bbox in scanfield
		m_LaserScanner.repeatPoint();
		delay = max(delay-10,0); // µs

		wfi();
	}

	// wait m_jumpDelay
	do
	{
		LOGTRACE("w");
		m_LaserScanner.repeatPoint();
		delay -= 10; // µs
	}
	while (delay > 0);
}

static inline // helper
int adjustDelay(int& delay)
{
	int d = min(delay,10);
	delay -= d;
	return d;
}

void ScanEngine::drawPolyLine(const Rect& bbox, uint count, std::function<Point()> readNextPoint)
{
	if (bbox.width() < 0 ||	bbox.height() < 0)  // allow zero width and height for "points"
	{
		throw CriticalError("Polyline: invalid bounding box");
	}

	if (count < 2)
	{
		throw CriticalError("Polyline: less than 2 points\n");
	}

	if (count > 10000)		// sanity
	{
		throw CriticalError("Polyline: more than 10000 points\n");
	}

	// get starting point and jump to it:
	jumpToPoint(bbox,readNextPoint());

	int on_delay = m_laserOnDelay_usec;		// remaining ON delay count down

	// draw lines from point to point
	// at start wait laserOnDelay before switching on the laser
	// between points wait polygonDelay

	LOGTRACE("\n");

	while (--count > 0)
	{
		// draw line:

		Point dest = readNextPoint();

		while (m_currentPoint != dest)
		{
			LOGTRACE("d");
			const pwr_t* pwr = powerOn - adjustDelay(on_delay); // (finally) all on
			m_currentPoint = movedTowardsForDraw(m_currentPoint,dest,m_markSpeed_mpss);
			m_LaserScanner.addPoint(m_currentPoint,pwr);
		}

		// if middle point (not last point)
		// wait polygonDelay:

		if (count != 1)
		{
			for (int mid_delay = m_markMiddleDelay_usec; mid_delay > 0; mid_delay -= 10)
			{
				LOGTRACE("w");
				const pwr_t* pwr = powerOn - adjustDelay(on_delay); // (finally) all on
				m_LaserScanner.addPoint(m_currentPoint,pwr);
			}
		}
	}

	// behind last point:
	// wait markDelay before jumping
	// and wait laserOffDelay before switching off the laser

	int off_delay = m_laserOffDelay_usec;	// remaining OFF delay count down

	for (int end_delay = m_markEndDelay_usec; end_delay > 0; end_delay -= 10)
	{
		LOGTRACE("w");
		const pwr_t* pwr = powerOff - adjustDelay(off_delay); // (finally) all off
		m_LaserScanner.addPoint(m_currentPoint,pwr);
	}
}

void ScanEngine::drawLine(const Rect& bbox, const Point& start, const Point& dest)
{
	// jump to start point:
	jumpToPoint(bbox,start);

	int on_delay = m_laserOnDelay_usec;		// remaining ON delay count down

	// draw line
	// at start wait laserOnDelay before switching on the laser

	while (m_currentPoint != dest)
	{
		LOGTRACE("d");
		const pwr_t* pwr = powerOn - adjustDelay(on_delay); // (finally) all on
		m_currentPoint = movedTowardsForDraw(m_currentPoint,dest,m_markSpeed_mpss);
		m_LaserScanner.addPoint(m_currentPoint,pwr);
	}

	// at end of line:
	// wait markDelay before jumping
	// and wait laserOffDelay before switching off the laser

	int off_delay = m_laserOffDelay_usec;	// remaining OFF delay count down

	for (int end_delay = m_markEndDelay_usec; end_delay > 0; end_delay -= 10)
	{
		LOGTRACE("w");
		const pwr_t* pwr = powerOff - adjustDelay(off_delay); // (finally) all off
		m_LaserScanner.addPoint(m_currentPoint,pwr);
	}
}

void ScanEngine::drawLine(const Point& start, const Point& dest)
{
	Rect bbox{max(start.y,dest.y),min(start.x,dest.x),min(start.y,dest.y),max(start.x,dest.x)};
	drawLine(bbox,start,dest);
}

void ScanEngine::drawRect(const Rect& bbox, uint i)
{
	Point p[5] = {bbox.top_left(), bbox.top_right(), bbox.bottom_right(), bbox.bottom_left()};
	drawPolyLine(bbox,5,[p,&i]()
	{
		return p[i++ & 3];
	});
}

void ScanEngine::drawEllipse(const Rect& bbox, uint steps, uint i)
{
	FLOAT fx=bbox.width()/2;
	FLOAT fy=bbox.height()/2;

	drawPolyLine(bbox,steps+1,[&fx,&fy,&i,&steps]()
	{
		FLOAT angle = 2*pi * i++ / steps;
		return Point(fx*cos(angle),fy*sin(angle));
	});
}

static inline void rotate_cw (Point& p, FLOAT sin, FLOAT cos)
{
	FLOAT px = p.x * cos + p.y * sin;
	FLOAT py = p.y * cos - p.x * sin;
	p.x = px;
	p.y = py;
}

#ifdef SCAN_ENGINE_IMPLEMENT_ANALOGUE_CLOCK_DEMO
void ScanEngine::drawClock (const Rect& bbox, uint plane_fast, uint plane_slow, uint plane_rounded, uint32 second)
{
	// copyright 2020-07-19 kio@little-bat.de
	// BSD 2-clause license

	assert(bbox.center() == Point(0,0));
	
	uint minute = second / 60;
	uint hour = minute / 60;
	second %= 60;
	minute %= 60;
	hour   %= 12;

	uint i;
	FLOAT rad;
	Point triangle[3];
	const FLOAT radius = bbox.width()/2;

	const FLOAT
		O_RING_DIA = 1.0,   // outer ring diameter
		I_RING_DIA = 0.05,  // clock axis diameter
		HM_O_DIA  = 0.95,   // hour markers outer diameter
		HM_I_DIA  = 0.80,   // hour markers inner diameter
		SH_LENGTH = 0.90, SH_I_DIA = -0.2,                   // seconds handle length and inner length
		MH_LENGTH = 0.70, MH_I_DIA = -0.15, MH_WIDTH = 0.05, // minute handle
		HH_LENGTH = 0.55, HH_I_DIA = -0.15, HH_WIDTH = 0.05; // hour handle

	// draw outer ring:
	setPlaneData(plane_rounded);
	drawEllipse(bbox * O_RING_DIA, 12*2, 3*2);   // 3 = start at 12 o'clock

	// draw full hour markers for 1 to 11 o'clock:
	
	setPlaneData(plane_fast);
	for (i=1; i<=11; i++)
	{
		rad = pi*2 / 12 * i;
		Point direction { radius * sin(rad), radius * cos(rad) };
		drawLine(direction * HM_O_DIA, direction * HM_I_DIA);
	}

	// draw hour marker for 12 o'clock: (a triangle)
	setPlaneData(plane_slow);
	FLOAT w = radius * (HM_O_DIA - HM_I_DIA) / 3;   // width/2 of the triangle
	triangle[0] = Point( 0, radius * HM_I_DIA);      // tip
	triangle[1] = Point(-w, radius * HM_O_DIA);
	triangle[2] = Point(+w, radius * HM_O_DIA);

	i=0; drawPolyLine(bbox, 4, [&i,&triangle](){ return triangle[i++ % 3]; }); // start drawing at tip

	// draw the handles starting and ending at the clock center to minimize jumps:

	// draw second handle:
	setPlaneData(plane_fast);
	rad = pi*2 / 60 * second;
	Point direction { radius * sin(rad), radius * cos(rad) };
	drawLine(bbox, direction * SH_I_DIA,  direction * SH_LENGTH);
	drawLine(bbox, direction * SH_LENGTH, direction * SH_I_DIA);

	// draw minute handle:
	setPlaneData(plane_fast);
	triangle[0] = Point( 0.0,      MH_LENGTH );		// tip
	triangle[1] = Point(-MH_WIDTH, MH_I_DIA);
	triangle[2] = Point(+MH_WIDTH, MH_I_DIA);

	rad = pi*2 / 60 * (minute + second/60.0);
	FLOAT sinus = sin(rad) * radius;
	FLOAT cosin = cos(rad) * radius;
	for (i=0; i<3; i++) { rotate_cw(triangle[i],sinus,cosin); }
	i=0; drawPolyLine(bbox,4,[&i,&triangle](){ return triangle[++i % 3]; }); // start drawing at base

	// draw hour handle:
	setPlaneData(plane_fast);
	triangle[0] = Point( 0.0,      HH_LENGTH );		// tip
	triangle[1] = Point(-HH_WIDTH, HH_I_DIA);
	triangle[2] = Point(+HH_WIDTH, HH_I_DIA);

	rad = pi*2 / 12 * (hour + minute/60.0);
	sinus = sin(rad) * radius;
	cosin = cos(rad) * radius;
	for (i=0; i<3; i++) { rotate_cw(triangle[i],sinus,cosin); }
	i=0; drawPolyLine(bbox,4,[&i,&triangle](){ return triangle[++i % 3]; }); // start drawing at base

	// draw inner ring: (axis)
	setPlaneData(plane_slow);
	drawEllipse(bbox * I_RING_DIA, 8);
}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_SHOW_CHECKER_BOARD
void ScanEngine::drawCheckerBoard (const Rect& bbox, uint count, uint flags)
{
	if (bbox.width() <= 0 || bbox.height() <= 0)
	{
		throw CriticalError("CheckerBoard: invalid bounding box");
	}

	if (count < 1)
	{
		throw CriticalError("CheckerBoard: invalid number of fields");
	}

	if (count > 32)		// sanity
	{
		throw CriticalError("CheckerBoard: too many fields");
	}

	bool show_axis = flags & 1;
	bool show_direction = flags & 2;

	for (uint i=0; i<=count; i++)
	{
		FLOAT y = bbox.bottom + bbox.height() * i / count;
		drawLine(Point(i & 1 ? bbox.right : bbox.left,y), Point(i & 1 ? bbox.left : bbox.right,y));
	}
	for (uint i=0; i<=count; i++)
	{
		FLOAT x = bbox.left + bbox.width() * i / count;
		drawLine(bbox, Point(x, i & 1 ? bbox.top : bbox.bottom), Point(x, i & 1 ? bbox.bottom : bbox.top));
	}

	if (show_axis)
	{
		//TODO
	}

	if (show_direction)
	{
		//TODO
	}
}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_SHOW_LISSAJOUS
void ScanEngine::drawLissajous (
		const Rect& bbox, FLOAT min_f_ratio, FLOAT max_f_ratio,
		uint32 steps_per_rot, uint32 rots_per_sweep )
{
	if (bbox.width() <= 0 || bbox.height() <= 0)
	{
		throw CriticalError("Lissajous: invalid bounding box");
	}
	FLOAT x_scale = bbox.width() / 2;
	FLOAT y_scale = bbox.height() / 2;
	Point center = bbox.center();

	steps_per_rot = minmax(4u, steps_per_rot, 1000u);		// steps for full circle in x direction
	rots_per_sweep = minmax(4u, rots_per_sweep, 1000000u);	// circles for full rotation min_fy/fx <-> max_fy/fx
	min_f_ratio = minmax(0.5, min_f_ratio, 10.05);			// min. fy/fx ratio
	max_f_ratio = minmax(min_f_ratio, max_f_ratio, 10.05);	// max. fy/fx ratio

	// draw infinite line:
	// don't bother with start and end

	FLOAT x_step = 2*pi / steps_per_rot;
	FLOAT x_angle = 0.0;
	FLOAT y_angle = 0.0;

	FLOAT f_ratio_center = sqrt(min_f_ratio * max_f_ratio);
	FLOAT f_ratio_angle_step = 2*pi / steps_per_rot / rots_per_sweep;
	FLOAT f_ratio_angle = 0.0;
	FLOAT f_ratio_scale = (max_f_ratio-min_f_ratio)/2;

	for (uint32 step=0; m_PrintServer.getAvail() == 0; step++)
	{
		if ((step & 0x01FF) == 0) update_stats();

		f_ratio_angle += f_ratio_angle_step;
		FLOAT f_ratio = f_ratio_center + sin(f_ratio_angle) * f_ratio_scale;

		x_angle += x_step;
		y_angle += x_step * f_ratio;

		FLOAT x = sin(x_angle) * x_scale;
		FLOAT y = cos(y_angle) * y_scale;

	    Point dest{center+Dist(x,y)};

		while (m_currentPoint != dest)
		{
			m_currentPoint = movedTowardsForDraw(m_currentPoint,dest,m_markSpeed_mpss);
			m_LaserScanner.addPoint(m_currentPoint,powerOn);
		}
	}
}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_SWING
void ScanEngine::showCalibrationSwing(const Rect& bbox, uint plane0, uint plane1)
{
	if (bbox.width() <= 0 || bbox.height() <= 0)
	{
		throw CriticalError("CalibrationSwing: invalid bounding box");
	}

	Point bl = bbox.bottom_left();
	Point br = bbox.bottom_right();
	Point tl = bbox.top_left();
	Point tr = bbox.top_right();
	Dist  w  = br-bl;

	// draw 2 horizontal lines using plane0

	setPlaneData(plane0);
	drawLine(tl,tr);
	drawLine(bl,br);

	// draw 3 vertical lines using plane1

	setPlaneData(plane1);
	drawLine(tl+w*0.8, bl+w*0.8);	// down
	drawLine(bl+w*0.7, tl+w*0.7);	// up
	drawLine(tl+w*0.6, bl+w*0.6);	// down

	// reduce height

	const FLOAT f = 0.5;
	tl.y *= f;
	bl.y *= f;
	tr.y *= f;
	br.y *= f;

	// draw 2 horizontal lines using plane0

	setPlaneData(plane0);
	drawLine(tl,tr);
	drawLine(bl,br);

	// draw 4 angled lines using plane1

	setPlaneData(plane1);
	Point p[3] = {bl+w*0.05, tl, bl};
	for(uint x=3; x<=9; x+=2)
	{
		p[1].x = p[0].x + w.dx*x/20;
		p[2].x = p[1].x + w.dx*x/20;

		uint i=0;
		drawPolyLine(bbox, 3, [p,&i](){return p[i++];});
		assert(i==3);
	}
}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_LASERONOFF
void ScanEngine::showCalibrationLaserOnOff(const Rect& bbox, uint plane0, uint plane1, uint min_delay, uint max_delay, uint delta_per_step)
{
	// show a "ladder" image with lines varying laser on and off delay from min_delay to max_delay
	// the steps are drawn in plane1, varying m_laserOnDelay and m_laserOffDelay
	// the jump and mark delay are set to maximum (200µs)
	// other parameters (e.g. jump and mark speed) are taken from plane data

	if (bbox.width() <= 0 || bbox.height() <= 0)
	{
		throw CriticalError("CalibrationLaserOnOff: invalid bounding box");
	}

	min_delay = minmax(0u,min_delay,200u);
	max_delay = minmax(min_delay,max_delay,200u);
	//num_steps = minmax(3u,num_steps,40u);
	delta_per_step = minmax(1u, delta_per_step, max_delay-min_delay);

	FLOAT t = bbox.top;
	FLOAT b = bbox.bottom;
	FLOAT l = bbox.left  * 0.25;	// shorter lines => less flicker
	FLOAT r = bbox.right * 0.25;

	setPlaneData(plane0);
	drawLine(Point(l,t),Point(l,b));
	drawLine(Point(r,b),Point(r,t));

	setPlaneData(plane1);
	m_jumpEndDelay_usec = 200;
	m_markEndDelay_usec = 200;

	t *= 0.9;
	b *= 0.9;

	//uint delta_per_step = (max_delay - min_delay) / num_steps;
	uint num_steps = (max_delay - min_delay) / delta_per_step;

	for (uint i = 0; i <= num_steps; i++)
	{
		m_laserOnDelay_usec  = min_delay + i * delta_per_step;
		m_laserOffDelay_usec = min_delay + i * delta_per_step;

		FLOAT y = t + i * (b-t) / num_steps;
		drawLine(Point(l,y), Point(r,y));
	}
}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_CIRCLE
void ScanEngine::showCalibrationCircle(const Rect& bbox, uint plane0, uint plane1)
{
	if (bbox.width() <= 0 || bbox.height() <= 0)
	{
		throw CriticalError("CalibrationCircle: invalid bounding box");
	}

	Rect box = bbox;
	for (uint i=0; i<3; i++)
	{
		setPlaneData(plane0);
		drawRect(box);

		setPlaneData(plane1);
		drawEllipse(box, 32-4*i, (32-4*i)*3/8);
		box = box * 0.666;
	}
}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_MOVE
void ScanEngine::drawCalibrationMove(const Rect& bbox)
{
	if (bbox.width() <= 0 || bbox.height() <= 0)
	{
		throw CriticalError("CalibrationMove: invalid bounding box");
	}

	TODO();
}
#endif

void ScanEngine::reset_stats() noexcept
{
	m_LaserScanner.reset_stats();
}

void ScanEngine::update_stats() noexcept
{
	m_PrintServer.updateOutputUnderruns(
			m_LaserScanner.m_output_points_underrun_with_laser_on_total,
			m_LaserScanner.m_output_points_underrun_with_laser_off_total);
	m_PrintServer.updateOutputPointBoundingBoxes(
			m_LaserScanner.m_bbox_of_output_points_with_laser_on,
			m_LaserScanner.m_bbox_of_output_points_with_laser_off);
	m_LaserScanner.reset_stats();
}

void ScanEngine::printOneLabel()
{
	LOGTRACE("[");

	// loop over all print commands in input data:
	while (true)

	switch (uint32 cmd = readUint())
	{

#ifdef SCAN_ENGINE_IMPLEMENT_ANALOGUE_CLOCK_DEMO
	case SharedMemory::SHOW_ANALOGUE_CLOCK:
	{
		LOGDEBUG("ScanEngine: show analogue clock");
		uint32 plane0 = readUint();
		uint32 plane1 = readUint();
		uint32 plane2 = readUint();
		uint32 dia_p100 = readUint(); if (dia_p100 > 100) throw CriticalError("analogue clock: ill. diameter");
		uint32 year  = readUint() % 1000 + 2000;
		uint32 month = readUint();    if (month-1 > 12) month = 0;  // 1 .. 12
		uint32 day   = readUint();    if (day-1 > 31) day = 0;      // 1 .. 31
		uint32 second = readUint() % (60*60*24);
		Rect box(m_LaserScanner.getScanField() * (dia_p100*0.01));

		uint32 start_count = interrupt_count;

		do
		{
			if (interrupt_count - start_count >= m_scannerFrequency)
			{
				start_count += m_scannerFrequency;
				if (++second == 60*60*24)
				{
					second = 0;
					if (++day > 31)
					{
						day = 1;
						if (++month > 12)
						{
							month = 1;
							year += 1;
						}
					}
				}
			}

			drawClock(box,plane0,plane1,plane2,second);
			update_stats();
		}
		while(m_PrintServer.getAvail() == 0);
		continue;
	}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_SHOW_LISSAJOUS
	case SharedMemory::SHOW_LISSAJOUS:
	{
		LOGDEBUG("ScanEngine: show lissajous");

		setPlaneData(readUint());
		Rect   box = readRect();
		FLOAT min = m_PrintServer.readDouble();
		FLOAT max = m_PrintServer.readDouble();
		uint32 spr = readUint();
		uint32 rps = readUint();
		drawLissajous(box,min,max,spr,rps);
		continue;
	}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_SHOW_CHECKER_BOARD
	case SharedMemory::SHOW_CHECKER_BOARD:
	{
		LOGDEBUG("ScanEngine: checker board");

		uint32 plane = readUint();
		Rect bbox = readRect();
		uint32 count = readUint();	// fields per row or col
		uint32 flags = readUint();

		do
		{	setPlaneData(plane);
			drawCheckerBoard(bbox,count,flags);
			update_stats();
			LOGTRACE("\n");
		}
		while(m_PrintServer.getAvail() == 0);
		continue;
	}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_SWING
	case SharedMemory::SHOW_CALIBRATION_SWING:
	{
		LOGDEBUG("ScanEngine: show calibration 'SWING'");

		int32 plane0 = readUint();
		int32 plane1 = readUint();
		Rect bbox = readRect();
		do
		{
			showCalibrationSwing(bbox,plane0,plane1);
			update_stats();
			LOGTRACE("\n");
		}
		while(m_PrintServer.getAvail() == 0);
		continue;
	}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_LASERONOFF
	case SharedMemory::SHOW_CALIBRATION_LASERONOFF:
	{
		LOGDEBUG("ScanEngine: show calibration 'LASER ON/OFF'");

		uint32 plane0 = readUint();
		uint32 plane1 = readUint();
		Rect bbox = readRect();
		uint32 min = readUint();
		uint32 max = readUint();
		uint32 step  = readUint();
		do
		{
			showCalibrationLaserOnOff(bbox,plane0,plane1,min,max,step);
			update_stats();
			LOGTRACE("\n");
		}
		while(m_PrintServer.getAvail() == 0);
		continue;
	}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_CIRCLE
	case SharedMemory::SHOW_CALIBRATION_CIRCLE:
	{
		LOGDEBUG("ScanEngine: show calibration 'CIRCLE'");

		int32 plane0 = readUint();
		int32 plane1 = readUint();
		Rect bbox = readRect();
		do
		{
			showCalibrationCircle(bbox,plane0,plane1);
			update_stats();
			LOGTRACE("\n");
		}
		while(m_PrintServer.getAvail() == 0);
		continue;
	}
#endif

#ifdef SCAN_ENGINE_IMPLEMENT_CALIBRATION_MOVE
	case SharedMemory::PRINT_CALIBRATION_MOVE:
	{
		LOGDEBUG("ScanEngine: print calibration 'MOVE'");

		setPlaneData(readUint());
		Rect bbox = readRect();
		drawCalibrationMove(bbox);
		update_stats();
		continue;
	}
#endif

	case SharedMemory::PRINT_POLYLINE:
	{
		//LOGTRACE("\n");
		//LOGTRACE("ScanEngine: polyline");

		setPlaneData(readUint());
		Rect bbox = readRect();
		uint32 count = readUint();
		drawPolyLine(bbox,count,[this](){return readPoint();});
		continue;
	}

	case SharedMemory::END_OF_JOB:
		LOGTRACE("]\n");
		LOGINFO("ScanEngine: end of job");
		return;

	default:
		if ((uint32)cmd < (uint32)SharedMemory::END_OF_JOB)
			throw CriticalError("unsupported command");
		else
			throw CriticalError("illegal command");
		break;
	}
}

void ScanEngine::handleOneJob()
{
	if (m_PrintServer.retainJob() == false) return;
	leds->setLedIdle(false);

	LOGINFO("ScanEngine: next job");
	leds->setLedJobActive(ON);

	// switch from scanner to label coordinate system:
	m_currentPoint = m_LaserScanner.labelForScanfield(m_currentPoint);

	try
	{
		reset_stats();
		printOneLabel();
		update_stats();
	}
	catch (FatalError&)
	{
		throw;
	}
	catch (std::exception& e) // incl. CriticalError
	{
		LOGERROR("ScanEngine: %s: aborting job",e.what());
	}

	// drain laser scanner
	// the encoder must not switch coordinate system for the next label before this label is finished!
	// => proper overrun detection in APU
	m_LaserScanner.waitEmpty();

	// switch from label to scanner coordinate system:
	m_currentPoint = m_LaserScanner.scanfieldForLabel(m_currentPoint);

	leds->setLedJobActive(OFF);
	m_PrintServer.releaseJob();
	m_LaserScanner.endOfJob();
}

__attribute__((noreturn))
void ScanEngine::run()
{
	//LOGINFO("ScanEngine: run");

	while (true)
	{
		handleOneJob();
		leds->setLedIdle(true);
		wfi();
	}
}

} /* namespace ScanEngine */













