/*
 * @file ScanEngine.h
 * @brief The Main Worker Class for the RPU
 *
 * The Main Worker Class for the RPU which provides the "Runner".
 *
 * @copyright redheads.de
 * @date 2020-03-29
 * @author guenter.woigk@redheads.de
 */

#ifndef SRC_SCANENGINE_H_
#define SRC_SCANENGINE_H_

#include "IPrintServer.h"
#include "QueuedLaserScanner.h"
#include "LaserPowerData.h"
#include <functional>
#include "rpuconfiguration.h"


namespace ScanEngine
{

/**
 * 	@brief The main class of this Library
 *
 * 	it provides a runner which reads data from a data source interface
 * 	and drives the laser power and scanner using a data output interface.
 * 	Messages are logged using macros which use a global static logger interface.
 */
class ScanEngine final
{
	/// Instance which provides the Input Data:
	IPrintServer& m_PrintServer;

	/// Instance which consumes the Output Data:
	QueuedLaserScanner& m_LaserScanner;

   // these settings are used by the LaserScanner
   // Indicate Laser Type and accordingly which laser parameters are used.
   //enum LaserType : uint32 { LaserTypeNone=0, LaserTypeCO2 } m_laserType = LaserTypeCO2;
   // Resolution of the scanner (16 or 18 bit)
   //int32 m_scannerResolutionBits = 16;

   uint32 m_scannerFrequency;     ///< Data Frequency of the Scanner
   FLOAT m_maxMarkSpeed_mpss;    ///< Hard Limit for Scanner Speed when Marking [m/SS]   (SS = Scanner Step)
   FLOAT m_minMarkSpeed_mpss;    ///< Hard Limit for Scanner Speed when Marking [m/SS]
   FLOAT m_maxJumpSpeed_mpss;    ///< Hard Limit for Scanner Speed when Moving with Laser Off [m/SS]
   FLOAT m_minJumpSpeed_mpss;    ///< Hard Limit for Scanner Speed when Moving with Laser Off [m/SS]

   /**
	* Minimum Delay after Jump [usec]
	* used to clamp the Jump Delay settings in plane data
	*/
   int32 m_minJumpDelay_usec = 0;

   /**
	* Minimum Delay after Mark [usec]
	* used to clamp the Mark Delay settings in plane data
	*/
   int32 m_minMarkDelay_usec = 0;

   /**
	* Minimum Delay at Vertice Point [usec]
	* used to clamp the Polygon Delay settings in plane data
	*/
   int32 m_minPolygonDelay_usec = 0;

   /**
	* Minimum Delay for Laser On after Line Start [usec]
	* used to clamp the Laser On Delay settings in plane data
	*/
   int32 m_minLaserOnDelay_usec = 0;

   /**
	* Minimum Delay for Laser Off after Mark [usec]
	* used to clamp the Laser Off Delay settings in plane data
	*/
   int32 m_minLaserOffDelay_usec = 0;

   /**
	* Maximum Delay after Jump [usec]
	* used to clamp the Jump Delay settings in plane data
	*/
   int32 m_maxJumpDelay_usec = 0;

   /**
	* Maximum Delay after Mark [usec]
	* used to clamp the Mark Delay settings in plane data
	*/
   int32 m_maxMarkDelay_usec = 0;

   /**
	* Maximum Delay at Vertice Point [usec]
	* used to clamp the Polygon Delay settings in plane data
	*/
   int32 m_maxPolygonDelay_usec = 0;

   /**
	* Maximum Delay for Laser On after Line Start [usec]
	* used to clamp the Laser On Delay settings in plane data
	*/
   int32 m_maxLaserOnDelay_usec = 0;

   /**
	* Maximum Delay for Laser Off after Mark [usec]
	* used to clamp the Laser Off Delay settings in plane data
	*/
   int32 m_maxLaserOffDelay_usec = 0;

	/**
	 * current scanner position (last point added to queue)
	 *
	 * while printing this is in label coodinates.
	 *
	 * between labels this is in scanner coordinates,
	 * because the label coordinate system will change.
	 */
	Point m_currentPoint = {0,0};

	// ---- plane data for current polyline  ----

    FLOAT m_markSpeed_mpss = -1;      ///< scan speed for marking (laser on) [m/SS]
    FLOAT m_jumpSpeed_mpss = -1;      ///< scan speed for jumping (laser off) [m/SS]
    int32 m_laserOnDelay_usec = -1;    ///< delay: start of mark vector -> laser on [µs]
    int32 m_laserOffDelay_usec = -1;   ///< delay: end of mark vector -> laser off [µs]
    int32 m_markEndDelay_usec = -1;    ///< delay: end of mark vector -> start of jump [µs]
    int32 m_jumpEndDelay_usec = -1;    ///< delay: end of jump vector -> start of marking [µs]
    int32 m_markMiddleDelay_usec = -1; ///< delay: at middle points of a polyline [µs]
    int32 m_spotSize_um = -1;          ///< spot size for filling algorithms [µm] (not used)

    /**
     * Set data members from plane data for next object to draw.
     *
     * @param planedata  plane data for drawing next object.
     */
	void setPlaneData(const PlaneData& planedata);

    /**
     * Set data members and setup m_LaserScanner from plane data.
     * The plane data is read from the m_PrintServer.
     *
     * @param plane  plane index for drawing next object.
     */
	void setPlaneData(uint plane);

	/**
	 * Jump to destination point.
	 *
	 * Wait for bounding box of next object to move into scan field
	 * and jump to the starting point of the next object using @ref m_jumpVelocity
	 * and wait the required delay @ref m_jumpDelay as set from plane data for this object.
	 *
	 * @param bbox  bounding box of next object
	 * @param dest  destination point
	 */
	void jumpToPoint(const Rect& bbox, const Point& dest);

	/**
	 * Draw a polygon line path.
	 *
	 * Wait for bounding box and jump to first point using @ref jumpToPoint().
	 * Reads the points by calling the std::function<Point()> @ref provider.
	 * Slice and draw lines using @ref m_markVelocity.
	 * At start delay laser power on by @ref m_laserOnDelay.
	 * At each middle point wait additional time @ref m_polygonDelay.
	 * After final point wait @ref m_markDelay before jumping
	 * and @ref m_laserOffDelay before switching off the laser.
	 *
	 * @param bbox        bounding box
	 * @param num_points  number of points, incl. start and end point
	 * @param provider    callable called to get the next point
	 */
	void drawPolyLine(const Rect& bbox, uint num_points, std::function<Point()> provider);

	/**
	 * Draw a single line
	 * The bounding box is calculated.
	 *
	 * @param start  start position
	 * @param end	 end position
	 */
	void drawLine(const Point& start, const Point& end);

	/**
	 * Draw a single line
	 *
	 * @param bbox   bounding box
	 * @param start  start position
	 * @param end	 end position
	 */
	void drawLine(const Rect& bbox, const Point& start, const Point& end);

	/**
	 * Draw a rectangle.
	 * The rectangle itself is used for it's bounding box.
	 * The starting point can be selected:
	 * - 0 = top left (default)
	 * - 1 = top right
	 * - 2 = bottom right
	 * - 3 = bottom left
	 *
	 * @param bbox         the rectangle
	 * @param start_index  start and end point
	 */
	void drawRect(const Rect& bbox, uint start_index = 0);

	/**
	 * Draw a circle or an ellipse.
	 * The Ellipse is drawn into the bounding box @ref bbox using @ref num_lines subdivisions.
	 * The starting point can be selected: Default is at 0° (x=max, y=0) and indexes of points increase CCW.
	 *
	 * @param bbox         the enclosing rectangle
	 * @param num_lines    number of subdivisions
	 * @param start_index  start and end point
	 */
	void drawEllipse(const Rect& bbox, uint num_lines, uint start_index = 0);

	/**
	 * Draw an analogue clock with the current date and time (if it was set properly).
	 *
	 * @param bbox        the bounding box of the clock
	 * @param plane0      fast lines
	 * @param plane1      slow line, text
	 * @param plane2      rounded lines. like fast, but with polygon_delay = 0.
	 */
	void drawClock(const Rect& bbox, uint plane0, uint plane1, uint plane2, uint32 second_of_day);

	/**
	 * Draw a checker board for calibration.
	 *
	 * @param bbox	 bounding box
	 * @param count  number of fileds per row and column
	 * @param flags	 draw axis arrows and direcion of movement (TODO)
	 */
	void drawCheckerBoard(const Rect& bbox, uint count, uint flags);

	/**
	 * Repeatedly draw a Lissajous figure for demonstration.
	 * The animation is terminated by the next drawing command or END_OF_JOB.
	 *
	 * @param bbox         bounding box
	 * @param fRatioMin    min. fy/fx ratio (>= 0.5)
	 * @param fRatioMax    max. fy/fx ratio (<= 10)
	 * @param stepsPerRot  steps for full circle in x direction (10..1000)
	 * @param rotsPerSweep circles for full rotation min_fy/fx <-> max_fy/fx  (10..100000)
	 */
	void drawLissajous (
			const Rect& bbox, FLOAT fRatioMin, FLOAT fRatioMax,
			uint32 stepsPerRot, uint32 rotsPerSweep );

	/**
	 * Draw a calibration test image
	 *
	 * This image draws vertical lines between 2 horizontal lines
	 * to show effective start and stop position and visible swing.
	 *
	 * @param bbox    bounding box
	 * @param plane0  plane for drawing helper lines
	 * @param plane1  plane for drawing lines for testing
	 */
	void showCalibrationSwing(const Rect& bbox, uint plane0, uint plane1);

	/**
	 * Draw a calibration test image
	 *
	 * This image shows the effective start and stop position depending on the programmed delays.
	 * The laser_on and laser_off delays for plane1 are modified programmatically.
	 *
	 * @param bbox    bounding box
	 * @param plane0  plane for drawing helper lines
	 * @param plane1  plane for drawing lines for testing
	 * @param min     min. value for laser ON and OFF delay
	 * @param max     max. value for laser ON and OFF delay
	 * @param step    step per line shown
	 */
	void showCalibrationLaserOnOff(const Rect& bbox, uint plane0, uint plane1, uint min, uint max, uint step);

	/**
	 * Draw a calibration test image
	 *
	 * This image shows the effect of scanner pull delay on circles.
	 * It draws a series of circles in squares in circles in squares and so on.
	 *
	 * @param bbox    bounding box
	 * @param plane0  plane for drawing squares (helper lines)
	 * @param plane1  plane for drawing circles (lines for testing)
	 */
	void showCalibrationCircle(const Rect& bbox, uint plane0, uint plane1);

	/**
	 * Print a calibration test image
	 *
	 * This image can be printed on the actual target or a test dummy.
	 * It prints crosshairs to check the ratio of movement encoder to scanfield size.
	 * It prints 5 crosshairs, 4 near the corners and one in the middle.
	 * The crosshairs will match if the ratio of movement encoder to scanfield match.
	 *
	 * Each crosshair consists of two parts:
	 *
	 * - The outer part printed early in the scanfield
	 * - The inner part printed late in the scanfield
	 *
	 * @param bbox    bounding box
	 */
	void drawCalibrationMove(const Rect& bbox);

	// helpers to read from source:
	Point readPoint() noexcept { return m_PrintServer.readPoint(); }
	int32 readInt()   noexcept { return m_PrintServer.readInt(); }
	uint32 readUint() noexcept { return m_PrintServer.readUint(); }
	Rect readRect()   noexcept { return m_PrintServer.readRect(); }

	/**
	 * Reset the collected statistics.
	 *
	 * The @ref QueuedLaserScanner counts Queue underruns
	 * and the effective bounding box of all scanned positions.
	 * These are forwarded to @ref SharedMemory::RpuStatus and can be printed on the debug console.
	 */
	void reset_stats() noexcept;

	/**
	 * Forward the collected statistics to @ref SharedMemory::RpuStatus.
	 *
	 * The @ref QueuedLaserScanner counts Queue underruns
	 * and the effective bounding box of all scanned positions.
	 * These are forwarded to @ref SharedMemory::RpuStatus and can be printed on the debug console.
	 */
	void update_stats() noexcept;

	/**
	 * Worker to read and print one SharedMemory buffer.
	 */
	void printOneLabel();

public:
	/**
	 * Constructor
	 *
	 * @param printserver  data source
	 * @param laserscanner data destination
	 * @param rpu_configuration rpu configuration blob from shared memory
	 */
	ScanEngine(IPrintServer& printserver, QueuedLaserScanner& laserscanner, const SharedMemory::RpuConfiguration& rpu_configuration);
	~ScanEngine() = default;   ///< D'tor

	/**
	 * 	Handle one print job.
	 *
	 * 	Test if a print job is pending and print it.
	 */
	void handleOneJob();

	/**
	 * 	Run ScanEngine infinitely.
	 *
	 * 	Does not return but may throw a @ref FatalError.
	 */
	void run() __attribute__((noreturn));
};

} /* namespace ScanEngine */

#endif /* SRC_SCANENGINE_H_ */







