#pragma once

#include "hardware/pio.h"
#include "cdefs.h"
#include "basic_geometry.h"
#include <functional>


class XY2
{
public:
    #define pio PIO_XY
    //static constexpr PIO pio      = XY_PIO;		impossible due to static cast addr -> ptr
    static constexpr uint sm_laser  = 0;
    static constexpr uint sm_clock  = 1;
    static constexpr uint sm_x      = 2;
    static constexpr uint sm_y      = 3;

    static constexpr uint pin_laser = PIN_XY2_LASER;
    static constexpr uint pin_clock = PIN_XY2_CLOCK;
    static constexpr uint pin_sync  = PIN_XY2_CLOCK + 2;
    static constexpr uint pin_x     = PIN_XY2_X;
    static constexpr uint pin_y     = PIN_XY2_Y;
    static constexpr uint pin_sync_xy = PIN_XY2_SYNC_XY;

    uint program_base_clock;	// in PIO program memory
    uint program_base_data;		// in PIO program memory
    uint program_base_laser;	// in PIO program memory

    static uint heart_beat_counter;
    static uint heart_beat_state;

    static int x0, y0;			// current scanner position


    XY2();
    static void start();

    static inline uint __tx_free() noexcept
	{
	    return 8 - pio_sm_get_tx_fifo_level(pio,sm_x);      // 8 = queue length if tx and rx fifo joined
	}

    static inline uint __tx_avail() noexcept
	{
	    return pio_sm_get_tx_fifo_level(pio,sm_x);
	}

    static inline void wait_tx_free () noexcept
    {
    	if (pio_sm_is_tx_fifo_full(pio,sm_x))
    	{
    		gpio_put(LED_IDLE,1);
	        while (pio_sm_is_tx_fifo_full(pio,sm_x)) {}
    		gpio_put(LED_IDLE,0);
    	}
    	else if (pio_sm_is_tx_fifo_empty(pio,sm_x))
    	{
    		gpio_xor_mask(1<<LED_PIN);    	 // visualize
    	}
    }

    static inline void send_data_blocking (int x, int y, bool laser_on=true)
    {
    	wait_tx_free();
        send_data(x,y,laser_on);
    }

    static inline void send_data_blocking (const Point& p, bool laser_on=true)
    {
    	wait_tx_free();
        send_data(int(p.x),int(p.y),laser_on);
    }

    static inline void send_data (int x, int y, bool laser_on=true)
    {
    	if (--heart_beat_counter == 0)
    	{
			heart_beat_counter = XY2_DATA_CLOCK / 2;	// => 1 Hz
			heart_beat_state = !heart_beat_state;
			gpio_put(LED_BEAT, heart_beat_state);
    	}

    	x = minmax(-0x8000, x, 0x7fff);
    	y = minmax(-0x7fff, y, 0x8000);

        pio_sm_put(pio, sm_x, 0x8000 + x);
        pio_sm_put(pio, sm_y, 0x8000 - y);		// y inverted for pos. y axis
        pio_sm_put(pio, sm_laser, laser_on?~0:0);
        //gpio_put(PIN_XY2_LASER,laser_on);
    }

	static void move_to (const Point& dest);
	static void draw_to (const Point& dest, FLOAT speed = SCANNER_MAX_STEP);
	static void drawLine (const Point& start, const Point& dest, FLOAT speed = SCANNER_MAX_STEP);
    static void drawRect (Rect rect, FLOAT speed = SCANNER_MAX_STEP);
	static void drawEllipse (const Rect& bbox, uint steps, uint i, FLOAT speed = SCANNER_MAX_STEP);
	static void drawPolyLine (uint count, std::function<Point()> readNextPoint, FLOAT speed = SCANNER_MAX_STEP);
	static void drawCheckerBoard (const Rect& bbox, uint count, FLOAT speed = SCANNER_MAX_STEP);
	static void drawClock (const Rect& bbox, uint32 second, FLOAT speed = SCANNER_MAX_STEP);
	static void drawLissajous (FLOAT width, FLOAT height, FLOAT min_f_ratio, FLOAT max_f_ratio,
							   uint32 steps_per_rot, uint32 rots_per_sweep, FLOAT speed = SCANNER_MAX_STEP);

private:
    void init_clock ();
    void init_data ();
    void init_laser ();
};


#undef pio	// PIO_XY
