#pragma once

#include <cmath>
#include <ostream>
#include "cdefs.h"

/**
 * Template for Data Type to Represent a Distance in 2-dim Space.
 */
template<typename T>
struct TDist
{
   T dx = 0, dy = 0;

   /// Default c'tor: create Dist with dx = dy = 0
   TDist()
   {
   }

   /**
    * c'tor with initial values
    *
    * @param dx
    * @param dy
    */
   TDist(T dx, T dy) :
		 dx(dx), dy(dy)
   {
   }

   /**
    * Multiply Distance with Factor.
    *
    * @param f   the factor
    * @return	 resulting distance
    */
   TDist operator* (T f) const
   {
	  return TDist(dx * f, dy * f);
   }

   /**
    * Compare two distances for Equality.
    *
    * @param lhs  the first distance
    * @param rhs  the other distance
    * @return     boolean result
    */
   friend bool operator== (const TDist& lhs, const TDist& rhs)
   {
	   return lhs.dx == rhs.dx && lhs.dy == rhs.dy;
   }

   /**
    * Calculate the Length of a Distance.
    *
    * @return  the length
    */
   T length() const noexcept { return sqrt(dx*dx+dy*dy); } // <cmath>

   /**
    * Divide Distance by Divisor.
    *
    * @param d  the divisor
    * @return   the resulting distance
    */
   TDist operator/ (T d) const { return TDist{dx/d,dy/d}; }

   /**
    * Add another Distances to this Distance.
    *
    * @param q  the distance to add
    * @return   the resulting distance
    */
   TDist operator+ (const TDist& q) const { return TDist{dx+q.dx,dy+q.dy}; }

   /**
    * Calculate normal vector of this distance.
    *
    * The normal vector has the same direction as this Distance but length = 1.
    *
    * @return  the normal vector
    */
   TDist normal () const { return (*this) / length(); }
};


/**
 * Template for Data Type to Represent a Point in 2-dim Space.
 */
template<typename T>
struct TPoint
{
   T x = 0, y = 0;

   /// Default c'tor with x = y = 0.
   TPoint()
   {
   }

   /**
    * Create point with initial values.
    *
    * @param x
    * @param y
    */
   TPoint(T x, T y) :
		 x(x), y(y)
   {
   }

   /**
    * Create point from another point with a different underlying type.
    *
    * @param q   the source point
    */
   template<typename Q>
   explicit TPoint(const Q &q) :
		 x(q.x), y(q.y)
   {
   }

   /**
    * Move Point by Distance: add Distance to Point.
    *
    * @param d  the Distance
    * @return   the resulting Point
    */
   TPoint operator+ (const TDist<T> &d) const
   {
	  return TPoint(x + d.dx, y + d.dy);
   }

   /**
    * Move Point by Distance: subtract Distance from Point.
    *
    * @param d  the Distance
    * @return   the resulting Point
    */
   TPoint operator- (const TDist<T>& d) const
   {
	  return TPoint(x - d.dx, y - d.dy);
   }

   /**
    * Calculate Distance between 2 Points.
    *
    * @param d   the other Point
    * @return    Distance between this and the other Point.
    */
   TDist<T> operator- (const TPoint &d) const
   {
	  return TDist<T>(x - d.x, y - d.y);
   }

   /**
    * Multiply Point with Factor.
    *
    * This scales the Image tho which this Point belongs
    * by this factor with the origin as the center.
    *
    * @param a   scaling factor
    * @return    the resulting Point position
    */
   TPoint operator* (T a) const
   {
	  return TPoint(x * a, y * a);
   }

   /**
    * Multiply Point by a Power of 2.
    *
    * Evtl. this is only possible for Points based on integer types.
    * Used to convert int16 to int32 based Points.
    *
    * @param n   shift distance (power of 2's exponent)
    * @return    the resulting Point
    */
   TPoint operator<< (int n) const
   {
	  return TPoint(x << n, y << n);
   }

   /**
    * Divide Point by a Power of 2.
    *
    * Evtl. this is only possible for Points based on integer types.
    * Used to convert int32 to int16 based Points.
    *
    * @param n   shift distance (power of 2's exponent)
    * @return    the resulting Point
    */
   TPoint operator>> (int n) const
   {
     return TPoint(x >> n, y >> n);
   }

   /**
    * Compare 2 Points for non-equality
    *
    * @param q   the other Point
    * @return    boolean result: true if points are not equal.
    */
   bool operator!= (const TPoint& q) const noexcept
   {
      return x!=q.x || y!=q.y;
   }

   /**
    * Compare 2 Points for Equality
    *
    * @param lhs  the first point
    * @param rhs  the other point
    * @return    boolean result: true if points are equal.
    */
   friend bool operator== (const TPoint& lhs, const TPoint& rhs)
   {
	   return lhs.x == rhs.x && lhs.y == rhs.y;
   }
};


/**
 * Template for Data Type to Represent a Rectangle in 2-dim Space.
 */
template<typename T>
struct TRect
{
   T top = 0, left = 0, bottom = 0, right = 0;

   /// Default c'tor: create empty Rect with all corners set to 0,0
   TRect()
   {
   }

   /**
    * Create Rect with initial values from 4 Coordinates.
    *
    * @param top
    * @param left
    * @param bottom
    * @param right
    */
   TRect(T top, T left, T bottom, T right) :
	   top(top), left(left), bottom(bottom), right(right)
   {
   }

   /**
    * Create Rect with initial values from 2 Points.
    *
    * @param topleft
    * @param bottomright
    */
   TRect(const TPoint<T> &topleft, const TPoint<T> &bottomright) :
	   top(topleft.y), left(topleft.x), bottom(bottomright.y), right(bottomright.x)
   {
   }

   /**
    * Move Rect by adding a Dist.
    *
    * @param d   the distance to move the Rect
    * @return    the resulting Rect
    */
   TRect operator+ (const TDist<T> &d) const noexcept
   {
	  return TRect(top + d.dy, left + d.dx, bottom + d.dy, right + d.dx);
   }

   /**
    * Move Rect by subtracting a Dist.
    *
    * @param d   the distance to move the the Rect
    * @return    the resulting Rect
    */
   TRect operator- (const TDist<T> &d) const noexcept
   {
	  return TRect(top - d.dy, left - d.dx, bottom - d.dy, right - d.dx);
   }

   /**
    * Scale Rect by a Factor.
    *
    * Scales the Rect by this factor with the origin as the center.
    *
    * @param f   the factor to multiply all corner Points with
    * @return    the resulting Rect
    */
   TRect operator* (T f) const noexcept		// scaled from origin
   {
	   return TRect( top * f, left * f, bottom * f, right * f);
   }

   /**
    * Calculate the Union of this and another Rect.
    *
    * @param q    the other Rect.
    */
   void uniteWith (const TRect& q)
   {
	   if (q.top > top) top = q.top;
	   if (q.bottom < bottom) bottom = q.bottom;
	   if (q.left < left) left = q.left;
	   if (q.right > right) right = q.right;
   }

   /**
    * Compare 2 Rects for Equality
    *
    * @param lhs  the first Rect
    * @param rhs  the other Rect
    * @return     boolean result: true if Rects are equal.
    */
   friend bool operator==(const TRect& lhs, const TRect& rhs)
   {
	   return lhs.left == rhs.left && lhs.top == rhs.top && lhs.right == rhs.right && lhs.bottom == rhs.bottom;
   }

   /// Get the Width of this Rect
   /// @return    the width of this Rect
   T width() const noexcept { return right - left; }

   /// Get the Height of this Rect
   /// @return    the height of this Rect
   T height() const noexcept { return top - bottom; }

   /// Get the bottom-left Point of this Rect
   /// @return    the bottom-left Point
   TPoint<T> bottom_left() const noexcept { return TPoint<T>{left,bottom}; }

   /// Get the bottom-right Point of this Rect
   /// @return    the bottom-right Point
   TPoint<T> bottom_right() const noexcept { return TPoint<T>{right,bottom}; }

   /// Get the top-left Point of this Rect
   /// @return    the top-left Point
   TPoint<T> top_left() const noexcept { return TPoint<T>{left,top}; }

   /// Get the top-right Point of this Rect
   /// @return    the top-right Point
   TPoint<T> top_right() const noexcept { return TPoint<T>{right,top}; }

   /// Get the Center of this Rect
   /// @return    the center Point
   TPoint<T> center() const noexcept { return TPoint<T>{(left+right)/2,(bottom+top)/2}; }

   /// Test whether this Rect is empty.
   /// A Rect is non-empty if width and height are > 0.
   /// @return    boolean result
   bool isEmpty() const noexcept { return right <= left || top <= bottom; }

   /// Test whether this Rect fully encloses another Rect.
   /// @param q   the other Rect
   /// @return    boolean result
   bool encloses(const TRect& q) const noexcept { return left<=q.left && right>=q.right && bottom<=q.bottom && top>=q.top; }

   /// Test whether a Point lies inside (or on an edge of) this Rect.
   /// @param p   the Point
   /// @return    boolean result
   bool contains(const TPoint<T>& p) const noexcept { return left<=p.x && right>=p.x && bottom<=p.y && top>=p.y; }

   /**
    * Calculate the Union of this Rect and a Point.
    * Grows the Rect so that it encloses the Point.
    *
    * @param p   the Point to include
    * @return    the resulting Rect.
    */
   TRect unitedWith (const TPoint<T>& p) const noexcept { return TRect(max(top,p.y),min(left,p.x),min(bottom,p.y),max(right,p.x)); }

   /**
    * Force a Point inside this Rect.
    * If the Point is inside this Rect then it is returned unmodified.
    * Else it is moved to the nearest boundary of this Rect.
    *
    * @param p   the Point to check
    * @return    the unmodified or adjusted Point
    */
   TPoint<T> forcedInside(const TPoint<T>& p) const noexcept { return TPoint<T>(minmax(left,p.x,right),minmax(bottom,p.y,top)); }

   /**
    * Grow this Rect so that it encloses the Point.
    * Calculates the Union of this Rect and a Point.
    *
    * Note: there is also method @ref unitedWith() which does not modify this Rect but returns the resulting Rect.
    *
    * @param p    the Point to add to this Rect.
    */
   void uniteWith (const TPoint<T>& p) noexcept
   {
	   if (p.y > top) top = p.y;
	   if (p.y < bottom) bottom = p.y;
	   if (p.x < left) left = p.x;
	   if (p.x > right) right = p.x;
   }

   /**
    * Grow this Rect at all sides by a certain distance.
    * If the Dist is negative then the Rect will shrink.
    *
    * @param d    the Distance to add.
    */
   void grow (T d) noexcept { left-=d; right+=d; bottom-=d; top+=d; }

   /**
    * Shrink this Rect at all sides by a certain distance.
    * If the Dist is negative then the Rect will grow.
    *
    * @param d    the Distance to add.
    */
   void shrink (T d) noexcept { grow(-d); }
};

/// The standard Point type used in the RPU, which uses double values
typedef struct TPoint<FLOAT> Point;
/// The standard Dist type used in the RPU, which uses double values
typedef struct TDist<FLOAT> Dist;
/// The standard Rect type used in the RPU, which uses double values
typedef struct TRect<FLOAT> Rect;

/// Point type which uses int32 values as used for the Scanner
typedef struct TPoint<int32> IntPoint;
/// Dist type which uses int32 values as used for the Scanner
typedef struct TDist<int32> IntDist;
/// Point type which uses int32 values as used for the Scanner
typedef struct TRect<int32> IntRect;


